@@ -10,20 +10,32 @@ import itertools
 import six
 import urwid
 from urwid_utils.palette import *
+# import urwid_readline
 from orderedattrdict import AttrDict
 
-from .datatable import *
-from .keymap import *
+# from .datatable import *
 from .listbox import ScrollingListBox
+from .keymap import *
+from .highlightable import HighlightableTextMixin
+from .autocomplete import AutoCompleteMixin
 
 class DropdownButton(urwid.Button):
 
+    text_attr = "dropdown_text"
+
     left_chars = u""
     right_chars = u""
 
-    def __init__(self, label, left_chars=None, right_chars=None):
+
+    def __init__(
+            self, label,
+            text_attr=None,
+            left_chars=None, right_chars=None
+    ):
 
         self.label_text = label
+        if text_attr:
+            self.text_attr = text_attr
         if left_chars:
             self.left_chars = left_chars
         if right_chars:
@@ -38,7 +50,7 @@ class DropdownButton(urwid.Button):
             ('weight', 1, self._label),
             (len(self.right_chars), self.button_right)
         ], dividechars=0)
-        self.set_label(("dropdown_text", self.label_text))
+        self.set_label((self.text_attr, self.label_text))
         super(urwid.Button, self).__init__(self.cols)
 
     @property
@@ -50,28 +62,44 @@ class DropdownButton(urwid.Button):
         return self.decoration_width + len(self.label_text)
 
 
-class DropdownItem(urwid.WidgetWrap):
+class DropdownItem(HighlightableTextMixin, urwid.WidgetWrap):
 
     signals = ["click"]
 
+    text_attr = "dropdown_text"
+    highlight_attr = "dropdown_highlight"
+    focused_attr = "dropdown_focused"
+
     def __init__(self, label, value,
-                 margin=0, left_chars=None, right_chars=None):
+                 margin=0,
+                 text_attr=None,
+                 focused_attr=None,
+                 highlight_attr=None,
+                 left_chars=None, right_chars=None):
 
         self.label_text = label
         self.value = value
         self.margin = margin
-        # self.button = urwid.Button(("dropdown_text", self.label_text))
+        if text_attr:
+            self.text_attr = text_attr
+        if focused_attr:
+            self.focused_attr = focused_attr
+        if highlight_attr:
+            self.highlight_attr = highlight_attr
         self.button = DropdownButton(
             self.label_text,
+            text_attr=self.text_attr,
             left_chars=left_chars, right_chars=right_chars
         )
-        # self.padding = urwid.Padding(self.button, width=("relative", 100),
-        #                              left=self.margin, right=self.margin)
-        self.padding = self.button
-        self.attr = urwid.AttrMap(self.padding, {None: "dropdown_text"})
+
+        self.padding = urwid.Padding(self.button, width=("relative", 100),
+                                     left=self.margin, right=self.margin)
+
+
+        self.attr = urwid.AttrMap(self.padding, {None: self.text_attr})
         self.attr.set_focus_map({
-            None: "dropdown_focused",
-            "dropdown_text": "dropdown_focused"
+            None: self.focused_attr,
+            self.text_attr: self.focused_attr
         })
         super(DropdownItem, self).__init__(self.attr)
         urwid.connect_signal(
@@ -80,6 +108,24 @@ class DropdownItem(urwid.WidgetWrap):
             lambda source: self._emit("click")
         )
 
+    @property
+    def highlight_source(self):
+        return self.label_text
+
+    @property
+    def highlightable_attr_normal(self):
+        return self.text_attr
+
+    @property
+    def highlightable_attr_highlight(self):
+        return self.highlight_attr
+
+    def on_highlight(self):
+        self.set_text(self.highlight_content)
+
+    def on_unhighlight(self):
+        self.set_text(self.highlight_source)
+
     @property
     def width(self):
         return self.button.width + 2*self.margin
@@ -101,164 +147,111 @@ class DropdownItem(urwid.WidgetWrap):
     def label(self):
         return self.button.label
 
-    def set_label(self, label):
-        logger.debug("set_label: " + repr(label) )
-        self.button.set_label(label)
-
-    def highlight_text(self, s, case_sensitive=False):
-
-        (a, b, c) = re.search(
-            r"(.*?)(%s)(.*)" %(s),
-            self.label_text,
-            re.IGNORECASE if not case_sensitive else 0
-        ).groups()
-        self.set_label([
-            ("dropdown_text", a),
-            ("dropdown_highlight", b),
-            ("dropdown_text", c),
-        ])
-
-    def unhighlight(self):
-        self.set_label(("dropdown_text", self.label_text))
-
-
-
-class AutoCompleteEdit(urwid.Edit):
-
-    signals = ["close"]
-
-    def keypress(self, size, key):
-        if key == "enter":
-            self._emit("close")
-        return super(AutoCompleteEdit, self).keypress(size, key)
-
-class AutoCompleteBar(urwid.WidgetWrap):
-
-    signals = ["change", "close"]
-    def __init__(self):
-
-        self.prompt = urwid.Text(("dropdown_prompt", "> "))
-        self.text = AutoCompleteEdit("")
-        # self.text.selectable = lambda x: False
-        self.cols = urwid.Columns([
-            (2, self.prompt),
-            ("weight", 1, self.text)
-        ], dividechars=0)
-        self.cols.focus_position = 1
-        self.filler = urwid.Filler(self.cols, valign="bottom")
-        urwid.connect_signal(self.text, "postchange", self.text_changed)
-        urwid.connect_signal(self.text, "close", lambda source: self._emit("close"))
-        super(AutoCompleteBar, self).__init__(self.filler)
-
-    def set_prompt(self, text):
-
-        self.prompt.set_text(("dropdown_prompt", text))
-
     def set_text(self, text):
-
-        self.text.set_edit_text(text)
-
-    def text_changed(self, source, text):
-        self._emit("change", text)
-
+        self.button.set_label(text)
 
 @keymapped()
-class DropdownDialog(urwid.WidgetWrap, KeymapMovementMixin):
+class DropdownDialog(AutoCompleteMixin, urwid.WidgetWrap, KeymapMovementMixin):
 
     signals = ["select", "close"]
 
+    text_attr = "dropdown_text"
+
     min_width = 4
 
+    label = None
+    border = None
+    scrollbar = False
+    margin = 0
+    max_height = None
+
     def __init__(
             self,
             drop_down,
             items,
             default=None,
-            label=None, border=False,
-            scrollbar=False,
-            auto_complete=False,
-            margin = 0,
+            label=None,
+            border=False,
+            margin = None,
+            scrollbar=None,
+            text_attr=None,
+            focused_attr=None,
+            prompt_attr=None,
             left_chars=None,
             right_chars=None,
             left_chars_top=None,
             rigth_chars_top=None,
-            max_height=10,
-            # keymap = {}
+            max_height=None,
+            keymap = {},
+            **kwargs
     ):
+
         self.drop_down = drop_down
         self.items = items
-        self.label = label
-        self.border = border
-        self.scrollbar = scrollbar
-        self.auto_complete = auto_complete
-        self.margin = margin
-        self.max_height = max_height
-
-        # self.KEYMAP = keymap
-
-        self.completing = False
-        self.complete_anywhere = False
-
+        if label is not None: self.label = label
+        if border is not None: self.border = border
+        if margin is not None: self.margin = margin
+        if scrollbar is not None: self.scrollbar = scrollbar
+        if text_attr:
+            self.text_attr = text_attr
+        if focused_attr:
+            self.focused_attr = focused_attr
+        if prompt_attr:
+            self.prompt_attr = prompt_attr
+        if max_height is not None: self.max_height = max_height
         self.selected_button = 0
         buttons = []
 
         buttons = [
                 DropdownItem(
                     label=l, value=v, margin=self.margin,
+                    text_attr=self.text_attr,
+                    focused_attr=self.focused_attr,
                     left_chars=left_chars,
                     right_chars=right_chars,
                 )
                 for l, v in self.items.items()
         ]
-        self.dropdown_buttons = ScrollingListBox(buttons, with_scrollbar=scrollbar)
+        self.dropdown_buttons = ScrollingListBox(
+            urwid.SimpleListWalker(buttons), with_scrollbar=scrollbar
+        )
 
         urwid.connect_signal(
             self.dropdown_buttons,
             'select',
-            lambda source, selection: self.select_button(selection)
+            lambda source, selection: self.on_complete_select(source)
         )
 
-        box_height = self.height -2 if self.border else self.height
-        self.box = urwid.BoxAdapter(self.dropdown_buttons, box_height)
-        self.fill = urwid.Filler(self.box)
         kwargs = {}
         if self.label is not None:
             kwargs["title"] = self.label
             kwargs["tlcorner"] = u"\N{BOX DRAWINGS LIGHT DOWN AND HORIZONTAL}"
             kwargs["trcorner"] = u"\N{BOX DRAWINGS LIGHT DOWN AND LEFT}"
 
-        w = self.fill
+        w = self.dropdown_buttons
         if self.border:
            w = urwid.LineBox(w, **kwargs)
 
-        if self.auto_complete:
-            self.auto_complete_bar = AutoCompleteBar()
-
-            urwid.connect_signal(self.auto_complete_bar,
-                                 "change",
-                                 lambda source, text: self.complete())
-
-            urwid.connect_signal(self.auto_complete_bar,
-                                 "close",
-                                 lambda source: self.complete_off())
-
         self.pile = urwid.Pile([
             ("weight", 1, w),
         ])
-        self.__super.__init__(self.pile)
-
+        super().__init__(self.pile)
 
     @property
-    def KEYMAP(self):
-        return self.drop_down.KEYMAP
+    def complete_container(self):
+        return self.pile
 
     @property
-    def filter_text(self):
-        return self.auto_complete_bar.text.get_text()[0]
+    def complete_container_pos(self):
+        return 1
 
-    @filter_text.setter
-    def filter_text(self, value):
-        return self.auto_complete_bar.set_text(value)
+    @property
+    def complete_body(self):
+        return self.body
+
+    @property
+    def complete_items(self):
+        return self.body
 
     @property
     def max_item_width(self):
@@ -303,108 +296,26 @@ class DropdownDialog(urwid.WidgetWrap, KeymapMovementMixin):
     def selection(self):
         return self.dropdown_buttons.selection
 
-    def select_button(self, button):
+    # def on_complete_select(self, pos, widget):
 
-        # logger.debug("select_button: %s" %(button))
-        label = button.label
-        value = button.value
-        self.selected_button = self.focus_position
-        self.complete_off()
-        self._emit("select", button)
-        self._emit("close")
+    #     # logger.debug("select_button: %s" %(button))
+    #     label = widget.label
+    #     value = widget.value
+    #     self.selected_button = self.focus_position
+    #     self.complete_off()
+    #     self._emit("select", widget)
+    #     self._emit("close")
 
     # def keypress(self, size, key):
-
-    #     raise Exception
-    #     logger.debug("DropdownDialog.keypress: %s" %(key))
-    #     if self.completing:
-    #         if key in ["enter", "up", "down"]:
-    #             self.complete_off()
-    #         else:
-    #             return key
-    #     else:
-    #         return super(DropdownDialog, self).keypress(size, key)
+    #     return super(DropdownDialog, self).keypress(size, key)
 
 
     @property
     def selected_value(self):
-
+        if not self.focus_position:
+            return None
         return self.body[self.focus_position].value
 
-    @keymap_command()
-    def complete_prefix(self):
-        self.complete_on()
-
-    @keymap_command()
-    def complete_substring(self):
-        self.complete_on(anywhere=True)
-
-    def complete_on(self, anywhere=False, case_sensitive=False):
-
-        if self.completing:
-            return
-        self.completing = True
-        # self.auto_complete_bar.set_prompt("> ")
-        # self.pile.focus_position = 1
-        self.show_bar()
-        if anywhere:
-            self.complete_anywhere = True
-        else:
-            self.complete_anywhere = False
-
-
-    @keymap_command()
-    def complete_off(self):
-
-        if not self.completing:
-            return
-        self.filter_text = ""
-
-        self.hide_bar()
-        self.completing = False
-
-    def complete(self, case_sensitive=False):
-
-        self[self.focus_position].unhighlight()
-        if not self.filter_text:
-            return
-
-        if case_sensitive:
-            g = lambda x: x
-        else:
-            g = lambda x: str(x).lower()
-
-        if self.complete_anywhere:
-            f = lambda x: g(self.filter_text) in g(x)
-        else:
-            f = lambda x: g(x).startswith(g(self.filter_text))
-
-        for i, w in enumerate(self.body):
-            if f(w):
-                self[i].highlight_text(self.filter_text)
-                self.focus_position = i
-                break
-
-    @keymap_command()
-    def cancel(self):
-        logger.debug("cancel")
-        self.focus_position = self.selected_button
-        self.close()
-
-    def close(self):
-        self._emit("close")
-
-    def show_bar(self):
-        self.pile.contents.append(
-            (self.auto_complete_bar, self.pile.options("given", 1))
-        )
-        self.pile.focus_position = 1
-
-    def hide_bar(self):
-        self[self.focus_position].unhighlight()
-        self.pile.focus_position = 0
-        del self.pile.contents[1]
-
 @keymapped()
 class Dropdown(urwid.PopUpLauncher):
     # Based in part on SelectOne widget from
@@ -412,38 +323,66 @@ class Dropdown(urwid.PopUpLauncher):
 
     signals = ["change"]
 
+    text_attr = "dropdown_text"
+    label_attr = "dropdown_label"
+    focused_attr = "dropdown_focused"
+    highlight_attr = "dropdown_highlight"
+    prompt_attr = "dropdown_prompt"
+
+    auto_complete = None
     label = None
     empty_label = u"\N{EMPTY SET}"
+    expanded = False
     margin = 0
 
     def __init__(
             self,
-            items = None,
-            label = None,
-            default = None,
-            border = False, scrollbar = False,
-            margin = None,
-            left_chars = None, right_chars = None,
-            left_chars_top = None, right_chars_top = None,
-            auto_complete = False,
-            keymap = {}
+            items=None,
+            label=None,
+            default=None,
+            expanded=None,
+            border=False, scrollbar=False,
+            margin=None,
+            text_attr=None,
+            label_attr=None,
+            focused_attr=None,
+            highlight_attr=None,
+            prompt_attr=None,
+            left_chars=None, right_chars=None,
+            left_chars_top=None, right_chars_top=None,
+            auto_complete=None,
+            max_height=10,
+            # keymap = {}
     ):
 
-        # raise Exception(self.KEYMAP_SCOPE)
         if items is not None:
             self._items = items
         if label is not None:
             self.label = label
+        if expanded is not None:
+            self.expanded = expanded
         self.default = default
 
         self.border = border
         self.scrollbar = scrollbar
-        self.auto_complete = auto_complete
+        if auto_complete is not None: self.auto_complete = auto_complete
+
         # self.keymap = keymap
 
         if margin:
             self.margin = margin
 
+        if text_attr:
+            self.text_attr = text_attr
+        if label_attr:
+            self.label_attr = label_attr
+        if focused_attr:
+            self.focused_attr = focused_attr
+        if highlight_attr:
+            self.highlight_attr = highlight_attr
+        if prompt_attr:
+            self.prompt_attr = prompt_attr
+
         if isinstance(self.items, list):
             if len(self.items):
                 if isinstance(self.items[0], tuple):
@@ -460,6 +399,9 @@ class Dropdown(urwid.PopUpLauncher):
         self.button = DropdownItem(
             u"", None,
             margin=self.margin,
+            text_attr=self.text_attr,
+            highlight_attr=self.highlight_attr,
+            focused_attr=self.focused_attr,
             left_chars = left_chars_top if left_chars_top else left_chars,
             right_chars = right_chars_top if right_chars_top else right_chars
         )
@@ -468,44 +410,57 @@ class Dropdown(urwid.PopUpLauncher):
             self,
             self._items,
             self.default,
-            label = self.label,
-            border = self.border,
-            margin = self.margin,
-            left_chars = left_chars,
-            right_chars = right_chars,
-            auto_complete = self.auto_complete,
-            scrollbar = scrollbar,
-            # keymap = self.KEYMAP
+            label=self.label,
+            border=self.border,
+            margin=self.margin,
+            text_attr=self.text_attr,
+            focused_attr=self.focused_attr,
+            prompt_attr=self.prompt_attr,
+            left_chars=left_chars,
+            right_chars=right_chars,
+            auto_complete=self.auto_complete,
+            scrollbar=scrollbar,
+            max_height=max_height,
+            # keymap=self.KEYMAP
         )
 
         urwid.connect_signal(
             self.pop_up,
             "select",
-            lambda souce, selection: self.select(selection)
+            lambda souce, pos, selection: self.select(selection)
         )
 
         urwid.connect_signal(
             self.pop_up,
             "close",
-            lambda button: self.close_pop_up()
+            lambda source: self.close_pop_up()
         )
 
         if self.default is not None:
             try:
-                self.select_value(self.default)
+                if isinstance(self.default, str):
+                    try:
+                        self.select_label(self.default)
+                    except ValueError:
+                        pass
+                else:
+                    raise StopIteration
             except StopIteration:
-                self.focus_position = 0
+                try:
+                    self.select_value(self.default)
+                except StopIteration:
+                    self.focus_position = 0
 
         if len(self):
             self.select(self.selection)
         else:
-            self.button.set_label(("dropdown_text", self.empty_label))
+            self.button.set_text((self.text_attr, self.empty_label))
 
         cols = [ (self.button_width, self.button) ]
 
         if self.label:
             cols[0:0] = [
-                ("pack", urwid.Text([("dropdown_label", "%s: " %(self.label))])),
+                ("pack", urwid.Text([(self.label_attr, "%s: " %(self.label))])),
             ]
         self.columns = urwid.Columns(cols, dividechars=0)
 
@@ -520,35 +475,37 @@ class Dropdown(urwid.PopUpLauncher):
             'click',
             lambda button: self.open_pop_up()
         )
+        if self.expanded:
+            self.open_pop_up()
 
     @classmethod
     def get_palette_entries(cls):
         return {
             "dropdown_text": PaletteEntry(
-                foreground = "light gray",
-                background = "dark blue",
-                foreground_high = "light gray",
-                background_high = "#003",
+                foreground="light gray",
+                background="dark blue",
+                foreground_high="light gray",
+                background_high="#003",
             ),
             "dropdown_focused": PaletteEntry(
-                foreground = "white",
-                background = "light blue",
-                foreground_high = "white",
-                background_high = "#009",
+                foreground="white",
+                background="light blue",
+                foreground_high="white",
+                background_high="#009",
             ),
             "dropdown_highlight": PaletteEntry(
-                foreground = "yellow",
-                background = "light blue",
-                foreground_high = "yellow",
-                background_high = "#009",
+                foreground="yellow",
+                background="light blue",
+                foreground_high="yellow",
+                background_high="#009",
             ),
             "dropdown_label": PaletteEntry(
-                foreground = "white",
-                background = "black"
+                foreground="white",
+                background="black"
             ),
             "dropdown_prompt": PaletteEntry(
-                foreground = "light blue",
-                background = "black"
+                foreground="light blue",
+                background="black"
             )
         }
 
@@ -602,6 +559,9 @@ class Dropdown(urwid.PopUpLauncher):
         height = self.pop_up.height + 1
         return height
 
+    def pack(self, size, focus=False):
+        return (self.width, self.height)
+
     @property
     def page_size(self):
         return self.pop_up.height
@@ -611,7 +571,7 @@ class Dropdown(urwid.PopUpLauncher):
         super(Dropdown, self).open_pop_up()
 
     def close_pop_up(self):
-        super(Dropdown, self).close_pop_up()
+        super().close_pop_up()
 
     def get_pop_up_parameters(self):
         return {'left': (len(self.label) + 2 if self.label else 0),
@@ -626,10 +586,38 @@ class Dropdown(urwid.PopUpLauncher):
 
     @focus_position.setter
     def focus_position(self, pos):
+        if pos == self.focus_position:
+            return
         # self.select_index(pos)
+        old_pos = self.focus_position
         self.pop_up.selected_button = self.pop_up.focus_position = pos
         self.select(self.selection)
 
+    @property
+    def items(self):
+        return self._items
+
+    @property
+    def selection(self):
+        return self.pop_up.selection
+
+    def select_label(self, label, case_sensitive=False):
+
+        old_value = self.value
+
+        f = lambda x: x
+        if not case_sensitive:
+            f = lambda x: x.lower()
+
+        index = next(itertools.dropwhile(
+                lambda x: f(x[1]) != f(label),
+                enumerate((self._items.keys())
+            )
+        ))[0]
+        self.focus_position = index
+
+
+
     @property
     def items(self):
         return self._items
@@ -641,18 +629,24 @@ class Dropdown(urwid.PopUpLauncher):
 
     def select_label(self, label, case_sensitive=False):
 
+        old_value = self.value
+
         f = lambda x: x
         if not case_sensitive:
-            f = lambda x: x.lower()
+            f = lambda x: x.lower() if isinstance(x, str) else x
 
-        index = next(itertools.dropwhile(
-                lambda x: f(x[1]) != f(label),
-                enumerate((self._items.keys())
-            )
-        ))[0]
+        try:
+            index = next(itertools.dropwhile(
+                    lambda x: f(x[1]) != f(label),
+                    enumerate((self._items.keys())
+                )
+            ))[0]
+        except StopIteration:
+            raise ValueError
         self.focus_position = index
 
 
+
     def select_value(self, value):
 
         index = next(itertools.dropwhile(
@@ -675,10 +669,41 @@ class Dropdown(urwid.PopUpLauncher):
     def selected_label(self):
         return self.selection.label
 
+    @selected_label.setter
+    def selected_label(self, label):
+        return self.select_label(label)
+
     @property
     def selected_value(self):
+        if not self.selection:
+            return None
         return self.selection.value
 
+    @selected_value.setter
+    def selected_value(self, value):
+        return self.select_value(value)
+
+    @property
+    def value(self):
+        return self.selected_value
+
+    @value.setter
+    def value(self, value):
+        old_value = self.value
+
+        # try to set by value.  if not found, try to set by label
+        try:
+            self.selected_value = value
+        except StopIteration:
+            self.selected_label = value
+
+    def cycle_prev(self):
+        self.cycle(-1)
+
+    def action(self):
+        pass
+
+    @keymap_command("cycle")
     def cycle(self, n):
         pos = self.focus_position + n
         if pos > len(self) - 1:
@@ -690,9 +715,11 @@ class Dropdown(urwid.PopUpLauncher):
 
     def select(self, button):
         logger.debug("select: %s" %(button))
-        self.button.set_label(("dropdown_text", button.label))
+        self.button.set_text((self.text_attr, button.label))
         self.pop_up.dropdown_buttons.listbox.set_focus_valign("top")
-        self._emit("change", button, button.value)
+        # if old_pos != pos:
+        self.action()
+        self._emit("change", self.selected_label, self.selected_value)
 
     # def set_items(self, items, selected_value):
     #     self._items = items
